#include <coap_comm.h>
#include <loc.h>
#include <zephyr/net/coap.h>
#include <modem/modem_key_mgmt.h>
#include <zephyr/net/tls_credentials.h>
#include <zephyr/net/socket.h>
#include <zephyr/random/random.h>

#include <zephyr/logging/log.h>

LOG_MODULE_DECLARE(nRF91DK_Tracker);

#define SEC_TAG 12
#define APP_COAP_VERSION 1
static int sock;
static struct sockaddr_storage server;
static uint8_t coap_buf[1280];
static uint16_t next_token;

int coap_credentials_init(void)
{
   LOG_INF("Configuting credentials for CoAP communication");
   int err;
   err = modem_key_mgmt_write(SEC_TAG, MODEM_KEY_MGMT_CRED_TYPE_IDENTITY, CONFIG_COAP_DEVICE_NAME,
                              strlen(CONFIG_COAP_DEVICE_NAME));
   if (err)
   {
      LOG_ERR("Failed to configure CoAP credentials: %d", err);
      return err;
   }
   err = modem_key_mgmt_write(SEC_TAG, MODEM_KEY_MGMT_CRED_TYPE_PSK, CONFIG_COAP_SERVER_PSK,
                              strlen(CONFIG_COAP_SERVER_PSK));
   if (err)
   {
      LOG_ERR("Failed to write identity: %d\n", err);
      return err;
   }

   LOG_INF("CoAP credentials configured successfully");
   return 0;
}

/* Resolve the CoAP server address */
int server_resolve(void)
{
   int err;
   struct addrinfo *result;
   struct addrinfo hints = {
       .ai_family = AF_INET,
       .ai_socktype = SOCK_DGRAM};
   char ipv4_addr[NET_IPV4_ADDR_LEN];

   LOG_INF("Resolving CoAP server address");
   err = getaddrinfo(CONFIG_COAP_SERVER_HOSTNAME, NULL, &hints, &result);
   if (err)
   {
      LOG_ERR("Failed to resolve CoAP server address: %d", err);
      return -EIO;
   }

   if (result == NULL)
   {
      LOG_ERR("No address found for CoAP server");
      return -ENOENT;
   }

   /* IPv4 Address. */
   struct sockaddr_in *server4 = ((struct sockaddr_in *)&server);

   server4->sin_addr.s_addr =
       ((struct sockaddr_in *)result->ai_addr)->sin_addr.s_addr;
   server4->sin_family = AF_INET;
   server4->sin_port = htons(CONFIG_COAP_SERVER_PORT);

   inet_ntop(AF_INET, &server4->sin_addr.s_addr, ipv4_addr,
             sizeof(ipv4_addr));
   LOG_INF("IPv4 Address found %s\n", ipv4_addr);

   /* Free the address. */
   freeaddrinfo(result);

   return 0;
}

/**@brief Initialize the CoAP client */
int client_init(void)
{
   int err;
   /* Create a DTLS socket */
   sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_DTLS_1_2);
   if (sock < 0)
   {
      LOG_ERR("Failed to create CoAP socket: %d.\n", errno);
      return -errno;
   }

   /* Set peer verification to be required */
   enum
   {
      NONE = 0,
      OPTIONAL = 1,
      REQUIRED = 2,
   };

   int verify = REQUIRED;

   err = setsockopt(sock, SOL_TLS, TLS_PEER_VERIFY, &verify, sizeof(verify));
   if (err)
   {
      LOG_ERR("Failed to setup peer verification, errno %d\n", errno);
      return -errno;
   }

   /* Set the DTLS hostname. */
   err = setsockopt(sock, SOL_TLS, TLS_HOSTNAME, CONFIG_COAP_SERVER_HOSTNAME,
                    strlen(CONFIG_COAP_SERVER_HOSTNAME));
   if (err)
   {
      LOG_ERR("Failed to setup TLS hostname (%s), errno %d\n",
              CONFIG_COAP_SERVER_HOSTNAME, errno);
      return -errno;
   }

   /* Set the credential security tag */
   sec_tag_t sec_tag_list[] = {SEC_TAG};
   err = setsockopt(sock, SOL_TLS, TLS_SEC_TAG_LIST, sec_tag_list,
                    sizeof(sec_tag_t) * ARRAY_SIZE(sec_tag_list));
   if (err)
   {
      LOG_ERR("Failed to setup socket security tag, errno %d\n", errno);
      return -errno;
   }

   err = connect(sock, (struct sockaddr *)&server,
                 sizeof(struct sockaddr_in));
   if (err < 0)
   {
      LOG_ERR("Connect failed : %d\n", errno);
      return -errno;
   }

   LOG_INF("Successfully connected to server");

   /*  Generate a random token after the socket is connected */
   next_token = sys_rand32_get();

   return 0;
}

int coap_comm_init(void)
{

   if (server_resolve() != 0)
   {
      LOG_ERR("Failed to resolve server name");
      return -1;
   }

   if (client_init() != 0)
   {
      LOG_ERR("Failed to initialize client");
      return -1;
   }

   return 0;
}

int coap_send_data(void)
{
   int err;
   struct coap_packet request;

   next_token = sys_rand32_get();

   /* Initialize the CoAP packet and append the resource path */
   err = coap_packet_init(&request, coap_buf, sizeof(coap_buf),
                          APP_COAP_VERSION, COAP_TYPE_NON_CON,
                          sizeof(next_token), (uint8_t *)&next_token,
                          COAP_METHOD_PUT, coap_next_id());
   if (err < 0)
   {
      LOG_ERR("Failed to create CoAP request, %d\n", err);
      return err;
   }

   err = coap_packet_append_option(&request, COAP_OPTION_URI_PATH,
                                   (uint8_t *)CONFIG_COAP_TX_RESOURCE,
                                   strlen(CONFIG_COAP_TX_RESOURCE));
   if (err < 0)
   {
      LOG_ERR("Failed to encode CoAP option, %d\n", err);
      return err;
   }

   /* Append the content format as plain text */
   const uint8_t text_plain = COAP_CONTENT_FORMAT_TEXT_PLAIN;
   err = coap_packet_append_option(&request, COAP_OPTION_CONTENT_FORMAT,
                                   &text_plain,
                                   sizeof(text_plain));
   if (err < 0)
   {
      LOG_ERR("Failed to encode CoAP option, %d\n", err);
      return err;
   }

   /* Add the payload to the message */
   err = coap_packet_append_payload_marker(&request);
   if (err < 0)
   {
      LOG_ERR("Failed to append payload marker, %d\n", err);
      return err;
   }

   err = coap_packet_append_payload(&request, (uint8_t *)location_buf, strlen(location_buf));
   if (err < 0)
   {
      LOG_ERR("Failed to append payload, %d\n", err);
      return err;
   }

   err = send(sock, request.data, request.offset, 0);
   if (err < 0)
   {
      LOG_ERR("Failed to send CoAP request, %d\n", errno);
      return -errno;
   }

   LOG_INF("CoAP PUT request sent: Token 0x%04x\n", next_token);

   return 0;
}